@@ -10,11 +10,12 @@ from pika.frame import Method
 from pika.spec import BasicProperties, Basic
 from vatis.asr_commons.config.logging import get_logger
 
-from . import constants
+from . import constants, parse
 from .connection import ReconnectingAMQPConnection, ConnectionState, ConnectionFactory
 from .exceptions import RetriesExceededException, ConnectionClosedException, NoConsumerException, \
     CorruptedBodyException, UnknownBodyTypeException
-from .model import Queue, Exchange, DEFAULT_EXCHANGE
+from .model import Queue, Exchange, DEFAULT_EXCHANGE, Message
+from .headers import *
 
 from typeguard import check_type
 
@@ -59,21 +60,17 @@ class Consumer:
             return NotImplemented
 
 
-@dataclasses.dataclass(frozen=True)
-class Message:
-    queue: str
-    exchange: str
-    body: bytes
-    headers: Dict[str, str]
+class RpcConsumer(Consumer):
+    callback: Callable[..., object]
 
 
 class MessageListener:
-    def on_message(self, message: Message) -> bool:
+    def on_message(self, message: Message) -> Tuple[bool, object]:
         """
         Raw message listener. It receives the message directly from the queue
         To not requeue the message, raise NoConsumerException
         :param message: received message
-        :return: acknowledged signal
+        :return: acknowledged signal and return body
         """
         pass
 
@@ -168,14 +165,38 @@ class ConsumerLoop:
 
     def _dispatch_message(self, channel: BlockingChannel, method: Basic.Deliver, properties: BasicProperties, body: bytes):
         if method.routing_key in self._queues_listener:
-            message: Message = Message(queue=method.routing_key, exchange=method.exchange, body=body,
-                                       headers=properties.headers if properties.headers is not None else {})
+            headers: Dict[str, str] = self._create_headers(properties)
+
+            message: Message = Message(queue=method.routing_key, exchange=method.exchange, body=body, headers=headers)
             listener: MessageListener = self._queues_listener[method.routing_key]
             ack: bool
             unconsumable: bool = False
 
             try:
-                ack = listener.on_message(message)
+                ack, response = listener.on_message(message)
+
+                if properties.reply_to is not None:
+                    try:
+                        if response is None:
+                            response = ''
+
+                        encoded_response, content_type = parse.encode(response)
+                        response_headers = {
+                            'Content-Type': content_type,
+                            'Content-Length': len(encoded_response),
+                            CORRELATION_ID_HEADER: properties.correlation_id
+                        }
+
+                        channel.basic_publish(exchange=constants.DEFAULT_EXCHANGE_NAME,
+                                              routing_key=properties.reply_to,
+                                              properties=BasicProperties(correlation_id=properties.correlation_id,
+                                                                         headers=response_headers,
+                                                                         delivery_mode=constants.DELIVERY_MODE_PERSISTENT,
+                                                                         content_type=content_type,
+                                                                         content_encoding=properties.content_encoding),
+                                              body=encoded_response)
+                    except Exception as e:
+                        logger.error(f'Response with correlation id {properties.correlation_id} could not be replied to {properties.reply_to}', e)
             except Exception as e:
                 logger.exception('Can\'t deliver message: %s', str(e))
                 ack = False
@@ -189,6 +210,21 @@ class ConsumerLoop:
             else:
                 channel.basic_nack(delivery_tag=method.delivery_tag, requeue=not unconsumable)
 
+    @staticmethod
+    def _create_headers(properties: BasicProperties) -> Dict[str, str]:
+        headers = {}
+
+        if properties.headers is not None:
+            headers.update(properties.headers)
+
+        if properties.correlation_id is not None:
+            headers[CORRELATION_ID_HEADER] = str(properties.correlation_id)
+
+        if properties.reply_to is not None:
+            headers[REPLY_TO_HEADER] = str(properties.reply_to)
+
+        return headers
+
     def listen_queue(self, queue: Queue, listener: MessageListener, exchange: Optional[Exchange] = DEFAULT_EXCHANGE):
         if self._closed:
             raise ConnectionClosedException()
@@ -312,13 +348,14 @@ class ConsumerManager(MessageListener):
                (c1 == c2 or \
                (c1.queue == c2.queue and c1.dtype == c2.dtype))
 
-    def on_message(self, message: Message) -> bool:
+    def on_message(self, message: Message) -> Tuple[bool, object]:
         from .parse import decode
 
         payload: Union[object, str, bytes] = decode(message)
 
         ack_message: bool = True
         delivered: bool = False
+        response: Optional[object] = None
 
         for consumer in self._queue_consumers[message.queue]:
             if consumer.dtype is None or _isinstance(payload, consumer.dtype):
@@ -330,7 +367,12 @@ class ConsumerManager(MessageListener):
                     if 'headers' in callback_signature.parameters:
                         additional_args['headers'] = message.headers.copy()
 
-                    ack: bool = consumer.callback(payload, **additional_args)
+                    ack: bool = True
+
+                    if isinstance(consumer, RpcConsumer):
+                        response = consumer.callback(payload, **additional_args)
+                    else:
+                        ack = consumer.callback(payload, **additional_args)
 
                     if not consumer.auto_ack:
                         ack_message = ack
@@ -339,10 +381,13 @@ class ConsumerManager(MessageListener):
                     if not consumer.auto_ack:
                         ack_message = False
 
+                    if isinstance(consumer, RpcConsumer):
+                        response = e
+
         if not delivered:
             raise NoConsumerException(message_type=type(payload))
 
-        return ack_message
+        return ack_message, response
 
     def close(self):
         self._consumer_loop.close()
