@@ -1,3 +1,6 @@
+from typing import Type, Optional
+
+
 class ConnectionClosedException(Exception):
     pass
 
@@ -11,7 +14,9 @@ class NoRouteFoundException(Exception):
 
 
 class NoConsumerException(Exception):
-    pass
+    def __init__(self, message_type: Optional[Type]):
+        super().__init__(f'Message type: {str(message_type)}')
+        self.message_type: Type = message_type
 
 
 class UnknownBodyTypeException(Exception):
